KEY_Z		= GLOBAL.KEY_Z
KEY_A		= GLOBAL.KEY_A
KEY_CTRL	= GLOBAL.KEY_CTRL
KEY_ALT	 	= GLOBAL.KEY_ALT
KEY_SHIFT 	= GLOBAL.KEY_SHIFT
string		= GLOBAL.string
tonumber	= GLOBAL.tonumber
GetPlayer	= GLOBAL.GetPlayer
TheInput	= GLOBAL.TheInput
TheSim		= GLOBAL.TheSim
TUNING 		= GLOBAL.TUNING
IsPaused	= GLOBAL.IsPaused
TheFrontEnd	= GLOBAL.TheFrontEnd
FindEntity 	= GLOBAL.FindEntity
GetWorld 	= GLOBAL.GetWorld
GetClock 	= GLOBAL.GetClock
SpawnPrefab 	= GLOBAL.SpawnPrefab
Vector3 	= GLOBAL.Vector3
FindWalkableOffset = GLOBAL.FindWalkableOffset
GetString 	= GLOBAL.GetString
COLLISION	= GLOBAL.COLLISION

local enableWallLabel = (GetModConfigData("Wall label")=="on")
local WallKey = GetModConfigData("Activate label on ALT+")

local bw = {
	hay 		= {mult = GetModConfigData("Hay wall"),			hp = TUNING.HAYWALL_HEALTH},
	wood 		= {mult = GetModConfigData("Wood wall"),		hp = TUNING.WOODWALL_HEALTH},
	stone		= {mult = GetModConfigData("Stone wall"),		hp = TUNING.STONEWALL_HEALTH},
	limestone	= {mult = GetModConfigData("Limestone wall (SW)"),	hp = 500},
	pomarble	= {mult = GetModConfigData("Marble wall (MOD)"),	hp = 600},
	nightmare	= {mult = GetModConfigData("Nightmare wall (MOD)"),	hp = 600},
	fa_lavawall	= {mult = GetModConfigData("Lava wall (MOD)"),		hp = 2000},
	ruins		= {mult = GetModConfigData("Ruins wall"),		hp = TUNING.RUINSWALL_HEALTH},
	}

local bwn = {}

----------------------------------------------------------------------------------------------------------

local showWallLabel = false
local BODYTEXTFONT = GLOBAL.BODYTEXTFONT

local function WallUpdateLabel(inst)
	if not enableWallLabel then return end
	if not inst.Label then return end
	if showWallLabel then
		local hc = math.floor(inst.components.health.currenthealth + 0.5)
		local hm = math.floor(inst.components.health:GetMaxHealth() + 0.5)
		local hp = math.floor(hc*100/hm + 0.5)
		local hstr = tostring(hc) .. "\n" .. tostring(hp).."\37"
		if hp > 50 then
			inst.Label:SetColour(12/255,236/255,154/255)
		elseif hp > 20 then
			inst.Label:SetColour(1,1,0)
		else
			inst.Label:SetColour(1,0,0)
		end
		inst.Label:SetText(hstr)
	end
	inst.Label:Enable(showWallLabel) 
end

local function WallAddLabel(inst)
	if not enableWallLabel then return end
	local l = inst.entity:AddLabel()
	inst.Label:SetFontSize(16)
	inst.Label:SetFont(BODYTEXTFONT)
	inst.Label:SetPos(0, 1, 0)
	inst.Label:SetText("")
	inst.Label:SetColour(12/255,236/255,154/255)
	inst:ListenForEvent( "bwlabel", function() WallUpdateLabel(inst) end , GetWorld())
end

----------------------------------------------------------------------------------------------

local function WallHealthStatus(Health,inst)
	if not enableWallLabel then return end
	inst:ListenForEvent("healthdelta", function(inst, data)
				if inst.components.health and inst:HasTag("wall") then
					local amount = (data.newpercent - data.oldpercent)*inst.components.health.maxhealth
					if (math.abs(amount) > 0.001) then WallUpdateLabel(inst) end
				end
			end)
end


local function BetterWallOnLoad(inst,data)
	inst.bwstart = true
	if inst.oldOnLoad and data then inst.oldOnLoad(inst,data) end
	if not (inst.components.bwdata and inst.components.bwdata.mult) then
		print("BetterWallOnLoad("..tostring(inst).."), bwdata not exists")
		return
	end
	local m = inst.components.bwdata.mult
	local hp = inst.components.bwdata.hp
	local h = inst.components.health.currenthealth
	if not (inst.bwname and bw[inst.bwname]) then
		print("BetterWallOnLoad("..tostring(inst).."), bwname is invalid "..tostring(inst.bwname))
		return
	end
	local d = bw[inst.bwname]
	local p = 0.5
	if not inst:HasTag("newwall") then
		if hp>0 then h=hp end
		if hp==0 then h=d.fullhp end
		p=h/(d.hp*m)
		if p<0 then p=0 end
		if p>1 then p=1 end
	end
	inst.components.bwdata.mult = d.mult
	inst.components.health.currenthealth = d.fullhp*p
	inst.components.health.maxhealth = d.fullhp
	inst.components.health:DoDelta(0)
	WallUpdateLabel(inst)
	inst.bwdone = true
	inst.bwstart = false
end

local function InitWall(i)
	local function checkinitfn() if i.initfn and i.initfn.Cancel then i.initfn:Cancel() i.initfn = nil end end
	if i.bwstart then return end
	if i and i.bwdone then checkinitfn() end
	if i and (not i.bwdone) and i.OnLoad then i:OnLoad() checkinitfn() end
end


local function WallToggleLabel()
	if not IsPaused() and not TheInput:IsKeyDown(KEY_CTRL) and TheInput:IsKeyDown(KEY_ALT) and not TheInput:IsKeyDown(KEY_SHIFT) then
		showWallLabel = not showWallLabel
		GetWorld():PushEvent("bwlabel",{}) 
	end
end

-------------------------------------------------------------------------------------------------------------------------

local function GetWallName(i)
	if not (i and i.prefab) then return "error" end
	if bwn[i.prefab] then return bwn[i.prefab] end -- supa-dupa caching
	local s = string.gsub(i.prefab, "_item$", "")
	s = string.gsub(s, "^wall_", "")
	s = string.gsub(s,"^mech_","")
	bwn[i.prefab] = s
--	print("wall name : "..tostring(i.name).." = "..tostring(s))
	return s
end

local function IsMech(i)
	if not (i and i.prefab) then return false end
	local s=string.gsub(i.prefab,"^mech_","")
	return (s~=i.prefab)
end

----------------------------------------------------------------------------------------------
-- from WallGate by _Q_

local function makeobstacle(inst)
	inst.Physics:SetCollisionGroup(COLLISION.OBSTACLES)
	inst.Physics:ClearCollisionMask()
	inst.Physics:SetMass(0)
	inst.Physics:CollidesWith(COLLISION.ITEMS)
	inst.Physics:CollidesWith(COLLISION.CHARACTERS)
	inst.Physics:SetActive(true)
	local ground = GetWorld()
	if ground then
		local pt = Point(inst.Transform:GetWorldPosition())
	    	ground.Pathfinder:AddWall(pt.x, pt.y, pt.z)
	end
end

local function openwall(inst)
	local name = nil
	if inst.entity:HasTag("grass") then 	name = "straw"
	elseif inst.entity:HasTag("wood") then 	name = "wood"
	elseif inst.entity:HasTag("stone") then	name = "stone"
	end

	local var = inst.components.health:GetPercent()
	if var <= 0 then
		inst.SoundEmitter:PlaySound("dontstarve/common/destroy_"..name)	
		inst.components.wallgates.isopen = false
		inst.AnimState:PushAnimation("0")
	elseif var <= .4 then
		inst.components.wallgates.isopen = true
		inst.AnimState:PlayAnimation("1_4")
		inst.SoundEmitter:PlaySound("dontstarve/common/place_structure_"..name)
		makeobstacle(inst)
	elseif var <= .5 then
		inst.components.wallgates.isopen = true
		inst.AnimState:PlayAnimation("1_2")
		inst.SoundEmitter:PlaySound("dontstarve/common/place_structure_"..name)
		makeobstacle(inst)
	elseif var <= .9 then
		inst.components.wallgates.isopen = true
		inst.AnimState:PlayAnimation("3_4")
		inst.SoundEmitter:PlaySound("dontstarve/common/place_structure_"..name)
		makeobstacle(inst)
	else
		inst.AnimState:PlayAnimation("1")
		inst.SoundEmitter:PlaySound("dontstarve/common/place_structure_"..name)
		inst.components.wallgates.isopen = true
		makeobstacle(inst)
	end
end

local function mech_ondeploywall(inst, pt, deployer)
	local mech = SpawnPrefab("mech_"..inst.bwname) 
	if mech then 
		pt = Vector3(math.floor(pt.x)+.5, 0, math.floor(pt.z)+.5)
		mech.Physics:SetCollides(false)
		mech.Physics:Teleport(pt.x, pt.y, pt.z) 
		mech.Physics:SetCollides(true)
		inst.components.stackable:Get():Remove()
		mech:AddTag("newwall")
		local ground = GetWorld()
		if ground then
			ground.Pathfinder:AddWall(pt.x, pt.y, pt.z)
	                openwall(mech)
		end
	end 		
end
-------------------------------------------------------------------------------------------------------

local function ondeploywall(inst, pt, deployer)
	local prefab = "wall_"..inst.bwname
	if inst.bwname == "fa_lavawall" then prefab = "fa_lavawall" end
	local wall = SpawnPrefab(prefab) 
		if wall then 
			pt = Vector3(math.floor(pt.x)+.5, 0, math.floor(pt.z)+.5)
			wall.Physics:SetCollides(false)
			wall.Physics:Teleport(pt.x, pt.y, pt.z) 
			wall.Physics:SetCollides(true)
			inst.components.stackable:Get():Remove()
			wall:AddTag("newwall")
		 	local ground = GetWorld()
		    	if ground then ground.Pathfinder:AddWall(pt.x, pt.y, pt.z) end
		end
end


local function BetterItemInit(inst)
	inst.bwname = GetWallName(inst)
--	print ("BetterItemInit : wname = "..tostring(inst.bwname)..", bw[bwname] = "..tostring(bw[inst.bwname]))
	
	if not (inst.bwname and bw[inst.bwname]) then
		print ("BetterItemInit("..tostring(inst)..") - unsupported prefab")
		return
	end
	if inst.components and inst.components.repairer then
		inst.components.repairer.healthrepairvalue = bw[inst.bwname].repair
	end
	if IsMech(inst) then
--		print ("BetterItemInit : IsMech")
		inst.components.deployable.ondeploy = mech_ondeploywall
	else
--		print ("BetterItemInit : IsWall")
		inst.components.deployable.ondeploy = ondeploywall
	end
end


local function BetterWallInit(inst)
	inst.bwname = GetWallName(inst)
--	print ("BetterWallInit : wname = "..tostring(inst.bwname)..", bw[bwname] = "..tostring(bw[inst.bwname]))
	if not (inst.bwname and bw[inst.bwname]) then
		print ("BetterWallInit("..tostring(i)..") - unsupported prefab")
		return
	end
	inst.oldOnLoad = inst.OnLoad
	inst.OnLoad = BetterWallOnLoad
	inst:AddComponent("bwdata")
	WallAddLabel(inst)
	inst.initfn = inst:DoPeriodicTask(0.2,InitWall,0.1)
end

for k,v in pairs(bw) do
	if v.mult and v.hp then
		bw[k].fullhp = v.mult*v.hp
		bw[k].repair = bw[k].fullhp/3
		if k == "fa_lavawall" then
			AddPrefabPostInit("fa_lavawall", BetterWallInit)
			AddPrefabPostInit("fa_lavawall_item", BetterItemInit)
		else
			AddPrefabPostInit("wall_"..k, BetterWallInit)
			AddPrefabPostInit("wall_"..k.."_item", BetterItemInit)
			AddPrefabPostInit("mech_"..k, BetterWallInit)
			AddPrefabPostInit("mech_"..k.."_item", BetterItemInit)
		end
	else
		print ("BetterWalls : WRONG DATA for "..k)
	end
end


if enableWallLabel and (WallKey > 0) then 
	local delta = KEY_A - string.byte("A")
	print ("BetterWalls : Alt+"..string.char(WallKey-delta))
	AddComponentPostInit("health",WallHealthStatus) 
	TheInput:AddKeyDownHandler(WallKey, WallToggleLabel)
end
